package com.datacomsolusindo.cpx_processor

import com.datacomsolusindo.cpx_shared_code.entity.*
import com.datacomsolusindo.cpx_shared_code.service.ApiService
import com.datacomsolusindo.cpx_shared_code.utility.*
import com.fasterxml.jackson.module.kotlin.readValue
import io.github.semutkecil.simplecriteria.FilterData
import jakarta.persistence.EntityManager
import jakarta.persistence.PersistenceContext
import jakarta.transaction.Transactional
import org.springframework.beans.factory.annotation.Autowired
import org.springframework.http.ResponseEntity
import org.springframework.stereotype.Service
import org.springframework.web.bind.annotation.PathVariable
import org.springframework.web.bind.annotation.PostMapping
import org.springframework.web.bind.annotation.RequestMapping
import org.springframework.web.bind.annotation.RestController
import java.time.LocalDateTime
import java.time.format.DateTimeFormatter
import java.util.*
import kotlin.time.measureTime
import kotlin.time.measureTimedValue


@Service
@Transactional
class ProcessorService {

    @Autowired
    lateinit var apiService: ApiService

    @Autowired
    lateinit var calculateCost2Service: CalculateCost2Service

    @PersistenceContext
    lateinit var em: EntityManager


    val dateFormatCriteria: DateTimeFormatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss")

    fun nullIfEmpty(data: String): String? {
        return if (data.trim().isEmpty()) {
            null
        } else {
            data
        }
    }

    @Transactional
    fun processData(raw: String) {
//        CoroutineScope(Dispatchers.IO).launch {
//        executor.execute {
        val logger = SimpleLogger.getLogger("process-${UUID.randomUUID()}")
        try {

            logger.info("process data $raw")
            val (dataSpl, measureSplit) = measureTimedValue { raw.split(",") }
            logger.info("split raw data ${measureSplit.inWholeMilliseconds}ms")
            val rawTransaction = RawCallTransaction()
            rawTransaction.startOfCall = dataSpl[3]
            rawTransaction.startOfAnswer = dataSpl[4]
            rawTransaction.endOfCall = dataSpl[5]
            rawTransaction.callerNumber = nullIfEmpty(dataSpl[13])
            rawTransaction.ownerPin = nullIfEmpty(dataSpl[15])
            rawTransaction.extTransferFrom = nullIfEmpty(dataSpl[17])
            rawTransaction.extTransferTo = nullIfEmpty(dataSpl[18])
            rawTransaction.pulse = nullIfEmpty(dataSpl[22])
//enum class TransferType { DIRECT, TRANSFER_FROM, TRANSFER_TO, TRANSFER_FROM_AND_TRANSFER_TO }
            //Kondisi transfer (D=direct, TF=transfer from, TT=transfer to, TE=transfer from and transfer to

            rawTransaction.transferType = when (dataSpl[19]) {
                "D" -> EnumData.TransferType.DIRECT
                "TF" -> EnumData.TransferType.TRANSFER_FROM
                "TT" -> EnumData.TransferType.TRANSFER_TO
                "TE" -> EnumData.TransferType.TRANSFER_FROM_AND_TRANSFER_TO
                else -> null
            }
            //Kondisi conference (S=single, C=Conference)
            rawTransaction.conferenceType = when (dataSpl[20]) {
                "S" -> EnumData.ConferenceType.SINGLE
                "C" -> EnumData.ConferenceType.CONFERENCE
                else -> null
            }
//            Jenis komunikasi (V=voice, D=data/fax, S=streaming video)
            rawTransaction.communicationType = when (dataSpl[20]) {
                "V" -> EnumData.CommunicationType.VOICE
                "D" -> EnumData.CommunicationType.DATA_FAX
                "S" -> EnumData.CommunicationType.STREAMING_VIDEO
                else -> null
            }

            //modifier
            val (data, measureData) = measureTimedValue {
                processModifier(
                    PreTransactionData(
                        pbxCode = dataSpl[1],
                        direction = dataSpl[2],
                        startOfCall = LocalDateTime.parse(dataSpl[3], dateFormatCriteria),
                        duration = dataSpl[7].toLong(),
                        dialedNumber = dataSpl[12],
                        callerNumber = dataSpl[13],
                        trunkCode = dataSpl[10],
                        ext = dataSpl[16],
                        extTransferFrom = dataSpl[17],
                        account = dataSpl[11],
                        accessNumber = dataSpl[8],
                        pin = dataSpl[14],
                        raw = raw
                    )
                )
            }
            logger.info("mapping data ${measureData.inWholeMilliseconds}ms")
            //count cost
            val (trunkData, measureTrunk) = measureTimedValue {
                apiService.findListAll(
                    Trunk::class.java, listOf(
                        "code",
                        "uid",
                        "area.uid",
                        "pbx.code",
                        "area.code",
                        "area.longitude",
                        "area.latitude",
                        "provider.code"
                    )
                ).firstOrNull { it["code"] == data.trunkCode && it["pbx.code"] == data.pbxCode }
            }

            logger.info("populate trunk data ${measureTrunk.inWholeMilliseconds}ms")

            val (pbxData, measurePbx) = measureTimedValue {
                apiService.findListAll(
                    Pbx::class.java, listOf(
                        "code", "uid", "provider.code", "area.code", "area.uid", "area.longitude", "area.latitude"
                    )
                ).firstOrNull { it["code"] == data.pbxCode }
            }

            logger.info("populate pbx data ${measurePbx.inWholeMilliseconds}ms")

            val (phoneUser, measurePhoneUser) = measureTimedValue {
                apiService.findListAll(
                    PhoneUserPbx::class.java, listOf(
                        "phoneUser.uid",
                        "phoneUser.id",
                        "phoneUser.counterCost",
                        "phoneUser.name",
                        "phoneUser.organization.uid",
                        "phoneUser.organization.id",
                        "phoneUser.organization.counterCost",
                        "phoneUser.costCenter.uid",
                        "phoneUser.costCenter.id",
                        "phoneUser.costCenter.counterCost",
                        "phoneUser.destinationGroup.uid",
                        "phoneUser.ratePage.name",
                        "phoneUser.ratePage.uid"
                    ), FilterData.and(
                        FilterData.filter("pbx.code", FilterData.FILTEROP.EQ, data.pbxCode), FilterData.or(
                            FilterData.filter("pin", FilterData.FILTEROP.EQ, data.pin ?: "#######"),
                            FilterData.filter("extension", FilterData.FILTEROP.EQ, data.ext ?: "#######")
                        )
                    )
                ).firstOrNull()
            }

            logger.info("populate phone user data ${measurePhoneUser.inWholeMilliseconds}ms")

            if (data.account != null) {
                val (accountData, measureAccount) = measureTimedValue {
                    apiService.findListAll(
                        Account::class.java, listOf("code", "uid")
                    ).firstOrNull { it["code"] == data.account }
                }
                logger.info("populate account data ${measureAccount.inWholeMilliseconds}ms")
                rawTransaction.accountUid = accountData?.get("uid") as String?
            }

            val callTo = calculateCost2Service.getCallTo(data.dialedNumber, logger)

            val callFrom = calculateCost2Service.getCallFrom(
                phoneUser, data.ext, data.pin, callTo?.zone != null, trunkData, pbxData, logger
            )

            rawTransaction.rawData = raw
            rawTransaction.pin = data.pin
            rawTransaction.extension = data.ext
            rawTransaction.number = data.dialedNumber
            rawTransaction.phoneUserUid = phoneUser?.get("phoneUser.uid") as String?
            rawTransaction.duration = data.duration
            rawTransaction.direction = data.direction
            rawTransaction.pbxUid = pbxData?.get("uid") as String?
            rawTransaction.trunkUid = trunkData?.get("uid") as String?
            rawTransaction.organizationUid = phoneUser?.get("phoneUser.organization.uid") as String?
            rawTransaction.costCenterUid = phoneUser?.get("phoneUser.costCenter.uid") as String?
            rawTransaction.destinationGroupUid = phoneUser?.get("phoneUser.destinationGroup.uid") as String?
            rawTransaction.rateUid = null
            rawTransaction.areaUid = callTo?.areaUid

            if (callTo != null && callFrom != null) {
                val zone = callTo.zone ?: if (callTo.domain == callFrom.domain) {
                    "LOC"
                } else {
                    "NDD"
                }

                val (zoneData, measureZone) = measureTimedValue {
                    apiService.findListAll(Zone::class.java, listOf(Zone::uid.name, Zone::code.name)).firstOrNull {
                        (it["code"] as String).lowercase(Locale.getDefault()) == zone.lowercase(
                            Locale.getDefault()
                        )
                    }
                }

                logger.info("populate zone data ${measureZone.inWholeMilliseconds}ms")

                rawTransaction.zoneUid = zoneData?.get("uid") as String?
                if (data.direction == "G") {
                    rawTransaction.cost = calculateCost2Service.calculateCost(
                        callTo, callFrom, zone, data.startOfCall, data.duration, logger
                    )

                    var cont = 1
                    val measureCalculateAdditionalCost = measureTime {
                        apiService.findListAll(
                            AdditionalCost::class.java, listOf(
                                AdditionalCost::sequence.name,
                                "crCostCenter.uid",
                                "crZone.uid",
                                "crTrunk.uid",
                                "crOrganization.uid",
                                "crAccount.uid",
                                "crPhoneUser.uid",
                                "crRatePage.uid",
                                AdditionalCost::crAppliedDate.name,
                                AdditionalCost::crDirection.name,
                                AdditionalCost::crExpiredDate.name,
                                AdditionalCost::continued.name,
                                AdditionalCost::percentage.name,
                                AdditionalCost::percentageBaseon.name,
                                AdditionalCost::target.name,
                                AdditionalCost::timeOffset.name,
                                AdditionalCost::value.name
                            ), sort = mutableListOf(Sort(AdditionalCost::sequence.name), Sort(AdditionalCost::id.name))
                        ).asSequence().filter {
                            it["value"] as Double != 0.0 || (it["percentage"] as Double != 0.0 && it["percentageBaseon"] != null)
                        }.filter {
                            it["crAppliedDate"] == null || (it["crAppliedDate"] as LocalDateTime).isBefore(data.startOfCall)
                        }.filter {
                            it["crExpiredDate"] == null || (it["crExpiredDate"] as LocalDateTime).isAfter(data.startOfCall)
                        }.filter {
                            it["crCostCenter.uid"] as String? == null || it["crCostCenter.uid"] == rawTransaction.costCenterUid
                        }.filter {
                            it["crZone.uid"] as String? == null || it["crZone.uid"] == rawTransaction.zoneUid
                        }.filter {
                            it["crTrunk.uid"] as String? == null || it["crTrunk.uid"] == rawTransaction.trunkUid
                        }.filter {
                            it["crOrganization.uid"] as String? == null || it["crOrganization.uid"] == rawTransaction.organizationUid
                        }.filter {
                            it["crAccount.uid"] as String? == null || it["crAccount.uid"] == rawTransaction.accountUid
                        }.filter {
                            it["crPhoneUser.uid"] as String? == null || it["crPhoneUser.uid"] == rawTransaction.phoneUserUid
                        }.filter {
                            it["crRatePage.uid"] as String? == null || it["crRatePage.uid"] == phoneUser?.get("phoneUser.ratePage.uid") as String?
                        }.filter {
                            it["crDirection"] as String? == null || (it["crDirection"] as String).contains(
                                rawTransaction.direction
                            ) // == rawTransaction.direction
                        }.toList().forEach {
                            if (cont == 1) {
                                val costAdd = if (it["value"] as Double != 0.0) {
                                    //count value
                                    it["value"] as Double
                                } else if (it["percentageBaseon"] != null) {
                                    //count pervent
                                    val basePercentage =
                                        (it["percentageBaseon"] as String).split(";").mapNotNull { bo ->
                                            try {
                                                EnumData.BaseOn.valueOf(bo.trim())
                                            } catch (e: Exception) {
                                                null
                                            }
                                        }.sumOf { ob ->
                                            when (ob) {
                                                EnumData.BaseOn.COST -> rawTransaction.cost
                                                EnumData.BaseOn.SERVICE -> rawTransaction.service
                                                EnumData.BaseOn.DISCOUNT -> rawTransaction.discount
                                                EnumData.BaseOn.TAX -> rawTransaction.tax
                                                EnumData.BaseOn.UNDEFINED -> 0.0
                                            }
                                        }

                                    basePercentage * (it["percentage"] as Double) / 100
                                } else {
                                    0.0
                                }

                                when (it["target"] as EnumData.BaseOn) {
                                    EnumData.BaseOn.COST -> rawTransaction.cost += costAdd
                                    EnumData.BaseOn.SERVICE -> rawTransaction.service += costAdd
                                    EnumData.BaseOn.DISCOUNT -> rawTransaction.discount += costAdd
                                    EnumData.BaseOn.TAX -> rawTransaction.tax += costAdd
                                    else -> {

                                    }
                                }
                            }
                            cont = (it["continued"] as Int)
                        }
                    }
                    logger.info("calculate additional cost ${measureCalculateAdditionalCost.inWholeMilliseconds}ms")

                }
                rawTransaction.phoneType = callFrom.phoneType?.ordinal ?: EnumData.PhoneType.UNDEFINED.ordinal
            }
            //add counter cost phone user
//            if (phoneUser?.get("phoneUser.id") != null) {
//                val currentCounterCost = phoneUser["phoneUser.counterCost"] as Double?
//                val counter = if (currentCounterCost == null) {
//                    rawTransaction.cost
//                } else {
//                    currentCounterCost + rawTransaction.cost
//                }
//                apiService.patch(
//                    PhoneUser::class.java,
//                    phoneUser["phoneUser.id"] as Long,
//                    mutableMapOf("counterCost" to counter)
//                )
//            }
//
//            if (phoneUser?.get("phoneUser.organization.id") != null) {
//                val currentCounterCost = phoneUser["phoneUser.organization.counterCost"] as Double?
//                val counter = if (currentCounterCost == null) {
//                    rawTransaction.cost
//                } else {
//                    currentCounterCost + rawTransaction.cost
//                }
//                apiService.patch(
//                    Organization::class.java,
//                    phoneUser["phoneUser.organization.id"] as Long,
//                    mutableMapOf("counterCost" to counter)
//                )
//            }
//
//            if (phoneUser?.get("phoneUser.costCenter.id") != null) {
//                val currentCounterCost = phoneUser["phoneUser.costCenter.counterCost"] as Double?
//                val counter = if (currentCounterCost == null) {
//                    rawTransaction.cost
//                } else {
//                    currentCounterCost + rawTransaction.cost
//                }
//                apiService.patch(
//                    CostCenter::class.java,
//                    phoneUser["phoneUser.costCenter.id"] as Long,
//                    mutableMapOf("counterCost" to counter)
//                )
//            }

            saveTransaction(rawTransaction)

        } catch (e: Exception) {
            logger.error("processor data error", e)
        }
    }

    @Transactional
    fun executeRawDataQuery(query: String, rawTransaction: RawCallTransaction, id: Long? = null) {
        val dirIndex = when (rawTransaction.direction) {
            "G" -> EnumData.Direction.OUTGOING.ordinal
            "C" -> EnumData.Direction.INCOMING.ordinal
            "I" -> EnumData.Direction.INTERNAL.ordinal
            else -> EnumData.Direction.UNDEFINED.ordinal
        }

        em.joinTransaction()
        val qb = em.createNativeQuery(query).setParameter("cost", rawTransaction.cost)
            .setParameter("currency", rawTransaction.currency).setParameter("direction", dirIndex)
            .setParameter("discount", rawTransaction.discount).setParameter("duration", rawTransaction.duration)
            .setParameter("extension", rawTransaction.extension).setParameter("number", rawTransaction.number)
            .setParameter("phoneType", rawTransaction.phoneType).setParameter("pin", rawTransaction.pin)
            .setParameter("rawData", rawTransaction.rawData).setParameter("rawData", rawTransaction.rawData)
            .setParameter("service", rawTransaction.service).setParameter("startOfCall", rawTransaction.startOfCall)
            .setParameter("tag", null).setParameter("tax", rawTransaction.tax).setParameter("tax", rawTransaction.tax)
            .setParameter("accountUid", rawTransaction.accountUid).setParameter("areaUid", rawTransaction.areaUid)
            .setParameter("costCenterUid", rawTransaction.costCenterUid)
            .setParameter("destinationGroupUid", rawTransaction.destinationGroupUid)
            .setParameter("organizationUid", rawTransaction.organizationUid)
            .setParameter("phoneUserUid", rawTransaction.phoneUserUid).setParameter("zoneUid", rawTransaction.zoneUid)
            .setParameter("rateUid", rawTransaction.rateUid).setParameter("trunkUid", rawTransaction.trunkUid)
            .setParameter("pbxUid", rawTransaction.pbxUid).setParameter("startOfAnswer", rawTransaction.startOfAnswer)
            .setParameter("endOfCall", rawTransaction.endOfCall)
            .setParameter("callerNumber", rawTransaction.callerNumber).setParameter("ownerPin", rawTransaction.ownerPin)
            .setParameter("extTransferFrom", rawTransaction.extTransferFrom)
            .setParameter("extTransferTo", rawTransaction.extTransferTo).setParameter("pulse", rawTransaction.pulse)
            .setParameter("transferType", rawTransaction.transferType?.ordinal)
            .setParameter("conferenceType", rawTransaction.conferenceType?.ordinal)
            .setParameter("communicationType", rawTransaction.communicationType?.ordinal)
        if (id != null) {
            qb.setParameter("id", id)
        }
        qb.executeUpdate()
        CacheData.update(Transaction::class.java)
    }

    @Transactional
    fun updateTransaction(rawTransaction: RawCallTransaction, id: Long) {
        val updateQuery =
            "UPDATE Customers " + "SET " + "cost = :cost, " + "currency = :currency, " + "direction = :direction, " + "discount = :discount, " + "duration = :duration, " + "extension = :extension, " + "number = :number, " + "phone_type = :phoneType, " + "pin = :pin, " + "raw_data = :rawData, " + "service = :service, " + "start_of_call = :startOfCall, " + "tag = :tag, " + "tax = :tax, " + "account_uid = :accountUid, " + "area_uid = :areaUid, " + "cost_center_uid = :costCenterUid, " + "destination_group_uid = :destinationGroupUid, " + "organization_uid = :organizationUid, " + "phone_user_uid = :phoneUserUid, " + "zone_uid = :zoneUid, " + "rate_id = :rateId, " + "trunk_uid = :trunkUid, " + "pbx_uid = :pbxUid, " + "start_of_answer = :startOfAnswer, " + "end_of_call = :endOfCall, " + "caller_number = :callerNumber, " + "owner_pin = :ownerPin, " + "ext_transfer_from = :extTransferFrom, " + "ext_transfer_to = :extTransferTo, " + "pulse = :pulse, " + "transfer_type = :transferType, " + "conference_type = :conferenceType, " + "communication_type= :communicationType" + "WHERE id = :id"
        executeRawDataQuery(updateQuery, rawTransaction, id)
    }

    @Transactional
    fun saveTransaction(rawTransaction: RawCallTransaction) {
        val insertQuery = "INSERT INTO calltransaction " + "(" +
//                "id, " +
                "cost, " + "currency, " + "direction, " + "discount, " + "duration, " + "extension, " + "number, " + "phone_type, " + "pin, " + "raw_data, " + "service, " + "start_of_call, " + "tag, " + "tax, " + "account_uid, " + "area_uid, " + "cost_center_uid, " + "destination_group_uid, " + "organization_uid, " + "phone_user_uid, " + "zone_uid, " + "rate_id, " + "trunk_uid, " + "pbx_uid, " + "start_of_answer, " + "end_of_call, " + "caller_number, " + "owner_pin, " + "ext_transfer_from, " + "ext_transfer_to, " + "pulse, " + "transfer_type, " + "conference_type, " + "communication_type" + ") VALUES (" +
//                "'1', " + //id
                ":cost, " + //cost
                ":currency, " + //currency
                ":direction, " + //direction
                ":discount, " + //discount
                ":duration, " + //duration
                ":extension, " + //extension
                ":number, " + //number
                ":phoneType, " + //phone_type
                ":pin, " + //pin
                ":rawData, " + //raw_data
                ":service, " + //service
                ":startOfCall, " + //start of call
                ":tag, " + // tag
                ":tax, " + // tax
                ":accountUid, " + // account_uid
                ":areaUid, " + // area_uid
                ":costCenterUid, " + // cost center uid
                ":destinationGroupUid, " + // destination group uid
                ":organizationUid, " + // organization_uid
                ":phoneUserUid, " + // phone user uid
                ":zoneUid, " + //zone uid
                ":rateUid, " + //rate_uid
                ":trunkUid, " + // trunk_uid
                ":pbxUid, " + //pbx_uid

                ":startOfAnswer, " + //pbx_uid
                ":endOfCall, " + //pbx_uid
                ":callerNumber, " + //pbx_uid
                ":ownerPin, " + //pbx_uid
                ":extTransferFrom, " + //pbx_uid
                ":extTransferTo, " + //pbx_uid
                ":pulse, " + //pbx_uid
                ":transferType, " + //pbx_uid
                ":conferenceType, " + //pbx_uid
                ":communicationType" + //pbx_uid
                ")"
        executeRawDataQuery(insertQuery, rawTransaction)
    }

    fun processModifier(data: PreTransactionData, lastOrder: Int = -100): PreTransactionData {
        val usedModifier = modifierLoader().filter { it.order > lastOrder }.firstOrNull {
            if (it.criteria == null) {
                true
            } else {
                val cProcess = it.criteria.sortedBy { c -> c.sequence }.map { c ->
                    val d: String? = when (c.attribute) {
                        ModifierCriteriaAttribute.PBX -> data.pbxCode
                        ModifierCriteriaAttribute.DIR -> data.direction
                        ModifierCriteriaAttribute.ACC -> data.account
                        ModifierCriteriaAttribute.ACN -> data.accessNumber
                        ModifierCriteriaAttribute.TRN -> data.trunkCode
                        ModifierCriteriaAttribute.NUM -> data.dialedNumber
                        ModifierCriteriaAttribute.EXT -> data.ext
                        ModifierCriteriaAttribute.SOC -> data.startOfCall.format(dateFormatCriteria)
                        ModifierCriteriaAttribute.DUR -> data.duration.toString()
                        ModifierCriteriaAttribute.PIN -> data.pin
                        ModifierCriteriaAttribute.NDG -> data.dialedNumber
                    }
                    criteriaMatcher(d, c.attribute, c.operator, c.value)
                }
                if (it.match == EnumData.CdrMatch.MATCH_ALL) {
                    cProcess.all { cb -> cb }
                } else {
                    cProcess.any { cb -> cb }
                }
            }
        }

        if (usedModifier == null) {
            return data
        }

        usedModifier.action?.sortedBy { it.sequence }?.filter { it.value.isNotEmpty() }?.forEach {
            when (it.attribute) {
                ModifierActionAttribute.DEF -> data.dialedNumber = data.dialedNumber.takeLast(
                    data.dialedNumber.length - (it.value.toIntOrNull() ?: 0)
                )

                ModifierActionAttribute.DEL -> data.dialedNumber = data.dialedNumber.take(
                    data.dialedNumber.length - (it.value.toIntOrNull() ?: 0)
                )

                ModifierActionAttribute.KEF -> data.dialedNumber = data.dialedNumber.take(
                    it.value.toIntOrNull() ?: data.dialedNumber.length
                )

                ModifierActionAttribute.KEL -> data.dialedNumber = data.dialedNumber.takeLast(
                    it.value.toIntOrNull() ?: data.dialedNumber.length
                )

                ModifierActionAttribute.INF -> data.dialedNumber = "${it.value}${data.dialedNumber}"
                ModifierActionAttribute.INL -> data.dialedNumber = "${data.dialedNumber}${it.value}"
                ModifierActionAttribute.CNU -> data.dialedNumber = it.value
                ModifierActionAttribute.CAN -> data.accessNumber = it.value
                ModifierActionAttribute.CTR -> data.trunkCode = it.value
                ModifierActionAttribute.CAC -> data.account = it.value
                ModifierActionAttribute.CPI -> data.pin = it.value
                ModifierActionAttribute.CDI -> data.direction = it.value
                ModifierActionAttribute.STE -> {
                    if (it.value == "Y") {
                        val trCOde = data.trunkCode
                        data.trunkCode = data.ext ?: ""
                        data.ext = trCOde
                    }
                }

                ModifierActionAttribute.ADU -> data.duration += (it.value.toLongOrNull() ?: 0)
                ModifierActionAttribute.RDU -> data.duration -= (it.value.toLongOrNull() ?: 0)
                ModifierActionAttribute.CDU -> data.duration = (it.value.toLongOrNull() ?: data.duration)
            }

        }

        return if (usedModifier.cont) {
            processModifier(data, usedModifier.order)
        } else {
            data
        }
    }


    private fun modifierLoader(): List<Modifier> {
        return apiService.findListPage(
            CdrModifier::class.java, listOf(
                CdrModifier::id.name,
                CdrModifier::criteria.name,
                CdrModifier::match.name,
                CdrModifier::action.name,
                CdrModifier::continued.name,
                CdrModifier::orderNumber.name,

                ), filter = FilterData.and(
                FilterData.filter(
                    "action", FilterData.FILTEROP.ISNOTNULL, ""
                ), FilterData.filter(
                    "match", FilterData.FILTEROP.NEQ, EnumData.CdrMatch.UNDEFINED.name
                )
            ), size = null, sort = mutableListOf(Sort("orderNumber"), Sort("id"))
        ).mapIndexed { i, it ->
            Modifier(
                order = i, match = it["match"] as EnumData.CdrMatch, criteria = if (it["criteria"] != null) {
                    Util.mapper.readValue<List<ModifierCriteria>>(it["criteria"].toString())
                } else {
                    null
                }, action = if (it["action"] != null) {
                    Util.mapper.readValue<List<ModifierAction>>(it["action"].toString())
                } else {
                    null
                }, cont = it["continued"] as Boolean
            )
        }.sortedBy { it.order }
    }

    fun criteriaMatcher(
        data: String?, attribute: ModifierCriteriaAttribute, operator: ModifierCriteriaOperator, value: String
    ): Boolean {
        return when (operator) {
            ModifierCriteriaOperator.CO -> data?.contains(value) == true
            ModifierCriteriaOperator.NC -> data?.contains(value) != true
            ModifierCriteriaOperator.IC -> value.split(",").map { it.trim() }.any { it == data }
            ModifierCriteriaOperator.XC -> !value.split(",").map { it.trim() }.any { it == data }
            ModifierCriteriaOperator.BG -> data?.startsWith(value) ?: false
            ModifierCriteriaOperator.ND -> data?.endsWith(value) ?: false
            ModifierCriteriaOperator.EQ -> data == value
            ModifierCriteriaOperator.NE -> data != value

            ModifierCriteriaOperator.LT -> if (data.isNullOrEmpty()) {
                false
            } else if (value.isEmpty()) {
                false
            } else {
                when (criteriaAttributeGroup(attribute)) {
                    "num" -> if (data.toDoubleOrNull() == null || value.toDoubleOrNull() == null) {
                        false
                    } else {
                        data.toDouble() < value.toDouble()
                    }

                    "dig" -> if (value.toIntOrNull() == null) {
                        false
                    } else {
                        data.length < value.toInt()
                    }

                    "dt" -> try {
                        LocalDateTime.parse(data, dateFormatCriteria)
                            .isBefore(LocalDateTime.parse(value, dateFormatCriteria))
                    } catch (e: Exception) {
                        false
                    }

                    else -> false
                }
            }


            ModifierCriteriaOperator.LE -> if (data.isNullOrEmpty()) {
                false
            } else if (value.isEmpty()) {
                false
            } else {
                when (criteriaAttributeGroup(attribute)) {
                    "num" -> if (data.toDoubleOrNull() == null || value.toDoubleOrNull() == null) {
                        false
                    } else {
                        data.toDouble() <= value.toDouble()
                    }

                    "dig" -> if (value.toIntOrNull() == null) {
                        false
                    } else {
                        data.length <= value.toInt()
                    }

                    "dt" -> try {
                        !LocalDateTime.parse(data, dateFormatCriteria)
                            .isAfter(LocalDateTime.parse(value, dateFormatCriteria))
                    } catch (e: Exception) {
                        false
                    }

                    else -> false
                }
            }

            ModifierCriteriaOperator.GT -> if (data.isNullOrEmpty()) {
                false
            } else if (value.isEmpty()) {
                false
            } else {
                when (criteriaAttributeGroup(attribute)) {
                    "num" -> if (data.toDoubleOrNull() == null || value.toDoubleOrNull() == null) {
                        false
                    } else {
                        data.toDouble() > value.toDouble()
                    }

                    "dig" -> if (value.toIntOrNull() == null) {
                        false
                    } else {
                        data.length > value.toInt()
                    }

                    "dt" -> try {
                        LocalDateTime.parse(data, dateFormatCriteria)
                            .isAfter(LocalDateTime.parse(value, dateFormatCriteria))
                    } catch (e: Exception) {
                        false
                    }

                    else -> false
                }
            }

            ModifierCriteriaOperator.GE -> if (data.isNullOrEmpty()) {
                false
            } else if (value.isEmpty()) {
                false
            } else {
                when (criteriaAttributeGroup(attribute)) {
                    "num" -> if (data.toDoubleOrNull() == null || value.toDoubleOrNull() == null) {
                        false
                    } else {
                        data.toDouble() >= value.toDouble()
                    }

                    "dig" -> if (value.toIntOrNull() == null) {
                        false
                    } else {
                        data.length >= value.toInt()
                    }

                    "dt" -> try {
                        !LocalDateTime.parse(data, dateFormatCriteria)
                            .isBefore(LocalDateTime.parse(value, dateFormatCriteria))
                    } catch (e: Exception) {
                        false
                    }

                    else -> false
                }
            }

            ModifierCriteriaOperator.BL -> data.isNullOrEmpty()
            ModifierCriteriaOperator.NB -> !(data.isNullOrEmpty())
        }
    }

    fun criteriaAttributeGroup(attribute: ModifierCriteriaAttribute): String {
        return when (attribute) {
            ModifierCriteriaAttribute.DIR, ModifierCriteriaAttribute.PBX, ModifierCriteriaAttribute.ACC, ModifierCriteriaAttribute.ACN, ModifierCriteriaAttribute.TRN, ModifierCriteriaAttribute.EXT, ModifierCriteriaAttribute.PIN, ModifierCriteriaAttribute.NUM -> "str"

            ModifierCriteriaAttribute.DUR -> "num"
            ModifierCriteriaAttribute.SOC -> "dt"
            ModifierCriteriaAttribute.NDG -> "dig"
        }
    }

}